@@ -19,6 +19,8 @@
 */
 
 /*
+| 2/6/2001 Patch from Laurent Caillat-Vallet <caillat@in2p3.fr>
+|	POP3/IMAP connections can be on non-standard port numbers.
 | 1/14/2001 Patch from Masaharu FUJITA <fujita@hpc.cs.ehime-u.ac.jp>
 |	The new mail krell animation for BSD is now a Daemon.
 | 11/30/2000 Patch from Adam Sampson <ats1@ukc.ac.uk>
@@ -27,16 +29,16 @@
 |	Added tooltip reporting of new mail counts for mailboxes checked via the
 |	Mail fetch/check program.
 |  7/4/2000 Patch from Michael Bell <mikeb@map.com>
-|	Added builtin POP3 and IMAP mail box checking routines.  Also added
-|	reporting the new mail counts for these to the tooltip popup.
+|	Added check_mail_pop3() and check_mail_imap() mail box checking routines.
+|	Also added reporting the new mail counts for these to the tooltip popup.
 |	Hajimu UMEMOTO added IPv6 awareness to these routines.
 |  4/25/2000 Patch from Stefan Schlott <stefan.schlott@student.uni-ulm.de>
 |	Added option: If there is new mail, the names of the mailboxes
 |	containing new mail can be displayed in a tooltip.
 |  3/16/2000 Patch from Stefan Schlott <stefan.schlott@student.uni-ulm.de>
 |	Added a check to avoid double counting old mail with status lines
-|	from Netscape and standard "Status" mail readers.  Also added a check
-|	to avoid counting mime attachments as messages, see is_multipart_mail()
+|	from Netscape and standard "Status" mail readers.  Also added
+|	is_multipart_mail() to avoid counting mime attachments as messages.
 |  2/24/2000 Patch from Darko Koruga <darko.koruga@hermes.si>
 |	Continuous mail animation option as long as there is new mail.
 | 12/16/1999 Patch from Patrick Morris <pmorris@starnixsolutions.com>
@@ -58,7 +60,7 @@
 #include "gkrellm_threads.h"
 
 #if defined(__linux__)
-#if defined(__GLIBC__) && ((__GLIBC__>2) || (__GLIBC__==2 && __GLIBC_MINOR__>=1))
+#if defined(__GLIBC__) && ((__GLIBC__>2)||(__GLIBC__==2 && __GLIBC_MINOR__>=1))
 #define HAVE_GETADDRINFO	1
 #endif
 #endif
@@ -88,7 +90,9 @@
 
 /* # of seconds to wait for a response from a POP3 or IMAP server
 */
-#define TCP_TIMEOUT  30
+#define TCP_TIMEOUT			30
+#define	DEFAULT_POP3_PORT	110
+#define	DEFAULT_IMAP_PORT	143
 
 
   /* A mailbox type has bits encoding how to check the mailbox (inline code
@@ -136,6 +140,7 @@ typedef struct
 	gchar	*username;		/* Account username			*/
 	gchar	*password;		/* Account password			*/
 	gchar	*server;		/* Mail Server (POP3/IMAP)	*/
+	gint	port;			/* Mail Server TCP Port */
 	gchar	*folder;		/* folder if server is IMAP*/
 	gint	busy;			/* TRUE if remote mailbox is being checked */
 	gint	type;			/* Type of mailbox: MBOX, POP3, IMAP, etc */
@@ -143,7 +148,7 @@ typedef struct
 	gboolean (*check_func)(); /* Mail check function to call. */
 	gpointer data;			/* For external mailboxes (in plugins) */
 
-	gint	mail_count;		/* Total number of messages in a mail box */
+	gint	mail_count;
 	gint	new_mail_count;
 	gint	old_mail_count;
 	gint	prev_mail_count,
@@ -238,14 +243,29 @@ free_mailbox(Mailbox *mbox)
 	g_free(mbox);
 	}
 
+
+static gchar *
+format_server_name(Mailbox *mbox)
+	{
+	static gchar	sname[128];
+
+	if (   (mbox->type == MBOX_POP3 && mbox->port != DEFAULT_POP3_PORT)
+		|| (mbox->type == MBOX_IMAP && mbox->port != DEFAULT_IMAP_PORT)
+	   )
+		snprintf(sname, sizeof(sname), "%s:%d", mbox->server, mbox->port);
+	else
+		snprintf(sname, sizeof(sname), "%s", mbox->server);
+	return sname;
+	}
+
 static gboolean
 format_remote_mbox_name(Mailbox *mbox, gchar *buf, size_t len)
 	{
 	if (mbox->folder)
 		snprintf(buf, len, "%s-%s@%s", mbox->username,
-			mbox->folder, mbox->server);
+			mbox->folder, format_server_name(mbox));
 	else if (mbox->server)
-		snprintf(buf, len, "%s@%s", mbox->username, mbox->server);
+		snprintf(buf, len, "%s@%s", mbox->username, format_server_name(mbox));
 	else if (mbox->username)
 		snprintf(buf, len, "%s", mbox->username);
 	else
@@ -341,11 +361,10 @@ is_From_line(Mailbox *mbox, gchar *buf)
 	}
 
 
-  /* Routine from Stefan Schlott,
-  | Check if this is a Content-Type-line. If it contains a boundary
-  | field, copy boundary string to buffer (including two leading and
-  | trailing dashes marking the end of a multipart mail) and return
-  | true. Otherwise, return false.
+  /* Check if this is a Content-Type-line. If it contains a boundary
+  |  field, copy boundary string to buffer (including two leading and
+  |  trailing dashes marking the end of a multipart mail) and return
+  |  true. Otherwise, return false.
   */
 static gint
 is_multipart_mail(gchar *buf, gchar *separator)
@@ -375,7 +394,8 @@ is_multipart_mail(gchar *buf, gchar *separator)
 				seplen = 0;
 				while (sepstart[seplen]!='"' && sepstart[seplen]>=32)
 					seplen++;
-				} else
+				}
+			else
 				{
 				seplen = 0;
 				while (sepstart[seplen]!=';' && sepstart[seplen]>32)
@@ -422,41 +442,71 @@ read_select(gint fd, gchar *buf, size_t size, time_t timeout)
 
 #if defined (GKRELLM_THREADS)
 
-  /* Read \r\n terminated lines from a remote IMAP or POP3 email server,
-  |  and verify the beginning of the line matches a string.
+  /* Read \r\n terminated lines from a remote IMAP or POP3 mail server,
   */
-static gboolean
-tcp_getline(Mailbox *mbox,  gint fd, gchar *match)
+static void
+tcp_getline(Mailbox *mbox,  gint fd)
 	{
 	gchar	buf[256];
 	gint	n;
-	gchar	c;
+	gchar	*s;
 
 	if (mbox->tcp_in)
 		mbox->tcp_in = g_string_truncate(mbox->tcp_in, 0);
 	else
 		mbox->tcp_in = g_string_new("");
-	while ((n = read_select(fd, buf, sizeof(buf) - 1, TCP_TIMEOUT)) > 0)
+	s = buf;;
+	while ((n = read_select(fd, s, 1, TCP_TIMEOUT)) > 0)
 		{
-		buf[n] = '\0';
-		g_string_append(mbox->tcp_in, buf);
-		c = mbox->tcp_in->str[mbox->tcp_in->len - 1];
-		if (c == '\r' || c == '\n')
+		*(s+1) = '\0';
+		if (*s++ == '\n')
 			break;
+		if (s >= buf + sizeof(buf) - 2)
+			{
+			g_string_append(mbox->tcp_in, buf);
+			s = buf;
+			}
 		}
+	if (s > buf)
+		g_string_append(mbox->tcp_in, buf);
+
 	if (GK.debug_level & DEBUG_MAIL)
 		{
 		if (n < 0)
 			perror("tcp_getline: ");
 		format_remote_mbox_name(mbox, buf, sizeof(buf));
-		printf("tcp_getline( %s )<%d>:%s\n", buf,
+		printf("server_response( %s )<%d>:%s\n", buf,
 			mbox->tcp_in->len, mbox->tcp_in->str);
 		}
+	}
+
+  /* Get a server response line and verify the beginning of the line
+  |  matches a string.
+  */
+static gboolean
+server_response(Mailbox *mbox, gint fd, gchar *match)
+	{
+	tcp_getline(mbox, fd);
 	return (!strncmp(match, mbox->tcp_in->str, strlen(match)) ? TRUE : FALSE);
 	}
 
+  /* Get a imap server completion result response for a tagged command.
+  |  Skip over any untagged responses the server may send.
+  */
+static gboolean
+imap_completion_result(Mailbox *mbox, gint fd, gchar *tag)
+	{
+	while (1)
+		{
+		tcp_getline(mbox, fd);
+		if (*(mbox->tcp_in->str) == '*')	/* untagged response */
+			continue;
+		return (!strncmp(tag, mbox->tcp_in->str, strlen(tag)) ? TRUE : FALSE);
+		}
+	}
+
 static void
-tcp_sendline(Mailbox *mbox, gint fd, gchar *line)
+server_command(Mailbox *mbox, gint fd, gchar *line)
 	{
 	gchar	buf[128];
 
@@ -464,7 +514,7 @@ tcp_sendline(Mailbox *mbox, gint fd, gchar *line)
 	if (GK.debug_level & DEBUG_MAIL)
 		{
 		format_remote_mbox_name(mbox, buf, sizeof(buf));
-		printf("tcp_sendline( %s ):%s", buf, line);
+		printf("server_command( %s ):%s", buf, line);
 		}
 	}
 
@@ -493,13 +543,6 @@ tcp_shutdown(gint fd, Mailbox *mbox, gchar *message)
 	return FALSE;
 	}
 
-
-  /* Added by Michael Bell <mikeb@map.com> 7/4/2000
-  |  check_mail_pop3 checks for mail on a pop3 server
-  |  mbox->username should be the username on the pop3 server
-  |  mbox->password should be the password on the pop3 server
-  |  mbox->server should be the fully qualified address of the pop3 server
-  */
 static gboolean
 check_mail_pop3(Mailbox *mbox)
 	{
@@ -508,17 +551,19 @@ check_mail_pop3(Mailbox *mbox)
 #ifdef HAVE_GETADDRINFO
 	gint 	rv	= 0;
 	struct addrinfo	hints, *res, *res0;
+	gchar	portnumber[6];
 #else
 	struct hostent	*addr;
 	struct sockaddr_in s;
 #endif
 
 #ifdef HAVE_GETADDRINFO
+	snprintf (portnumber, sizeof(portnumber), "%d", mbox->port);
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_family = PF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_protocol = 0;
-	if ((rv = getaddrinfo(mbox->server, "110", &hints, &res0)) != 0)
+	if ((rv = getaddrinfo(mbox->server, portnumber, &hints, &res0)) != 0)
 		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
 
 	for (res = res0; res; res = res->ai_next)
@@ -546,7 +591,7 @@ check_mail_pop3(Mailbox *mbox)
 		return FALSE;
 
 	s.sin_family = AF_INET;
-	s.sin_port = htons (110);
+	s.sin_port = htons (mbox->port);
 
 	if (connect (fd, (struct sockaddr *)&s, sizeof (s)) < 0)
 		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
@@ -554,12 +599,12 @@ check_mail_pop3(Mailbox *mbox)
 
 	/* Is the machine we are connected to really a POP3 server?
 	*/
-	if (! tcp_getline(mbox, fd, "+OK"))
+	if (! server_response(mbox, fd, "+OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[1]);
 
 	snprintf (line, sizeof (line), "USER %s\r\n", mbox->username);
-	tcp_sendline(mbox, fd, line);
-	if (! tcp_getline(mbox, fd, "+OK"))
+	server_command(mbox, fd, line);
+	if (! server_response(mbox, fd, "+OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[2]);
 
 	snprintf (line, sizeof (line), "PASS %s\r\n", mbox->password);
@@ -570,22 +615,22 @@ check_mail_pop3(Mailbox *mbox)
 
 		hide_password(mbox, line, 5);
 		format_remote_mbox_name(mbox, buf, sizeof(buf));
-		printf("tcp_sendline( %s ):%s", buf, line);
+		printf("server_command( %s ):%s", buf, line);
 		}
-	if (! tcp_getline(mbox, fd, "+OK"))
+	if (! server_response(mbox, fd, "+OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[3]);
 
-	tcp_sendline(mbox, fd, "STAT\r\n");
-	if (! tcp_getline(mbox, fd, "+OK"))
+	server_command(mbox, fd, "STAT\r\n");
+	if (! server_response(mbox, fd, "+OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[4]);
 
 	sscanf(mbox->tcp_in->str, "+OK %d", &mbox->mail_count);
    	snprintf (line, sizeof (line), "UIDL %d\r\n", mbox->mail_count);
-	tcp_sendline(mbox, fd, line);
-	if (! tcp_getline(mbox, fd, "+OK"))
+	server_command(mbox, fd, line);
+	if (! server_response(mbox, fd, "+OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[5]);
 
-	tcp_sendline(mbox, fd, "QUIT\r\n");
+	server_command(mbox, fd, "QUIT\r\n");
 	close (fd);
 
 	/* Set the new_mail_count only if the UIDL is changed to avoid
@@ -598,13 +643,6 @@ check_mail_pop3(Mailbox *mbox)
 	return TRUE;
 	}
  
-  /* Added by Michael Bell <mikeb@map.com> 7/4/2000
-  |  check_mail_imap checks for mail on a imap server
-  |  mbox->username should be the username on the imap server
-  |  mbox->password should be the password on the imap server
-  |  mbox->server should be the fully qualified address of the imap server
-  |  mbox->folder should be the folder on the imap server
-  */
 static gboolean
 check_mail_imap(Mailbox *mbox)
 	{
@@ -612,7 +650,9 @@ check_mail_imap(Mailbox *mbox)
 	gint	messages	= 0;
 	gint	unseen		= 0;
 	gchar	line[128], *ss;
+
 #ifdef HAVE_GETADDRINFO
+	gchar	portnumber[6];
 	gint	rv	= 0;
 	struct addrinfo hints, *res, *res0;
 #else
@@ -621,11 +661,12 @@ check_mail_imap(Mailbox *mbox)
 #endif
 
 #ifdef HAVE_GETADDRINFO
+	snprintf (portnumber, sizeof(portnumber), "%d", mbox->port);
 	memset(&hints, 0, sizeof(hints));
 	hints.ai_family = PF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_protocol = 0;
-	if ((rv = getaddrinfo(mbox->server, "143", &hints, &res0)) != 0)
+	if ((rv = getaddrinfo(mbox->server, portnumber, &hints, &res0)) != 0)
 		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
 
 	for (res = res0; res; res = res->ai_next)
@@ -653,7 +694,7 @@ check_mail_imap(Mailbox *mbox)
 		return FALSE;
 
 	s.sin_family = AF_INET;
-	s.sin_port = htons (143);
+	s.sin_port = htons (mbox->port);
 
 	if (connect (fd, (struct sockaddr *)&s, sizeof (s)) < 0)
 		return tcp_shutdown(fd, mbox, tcp_error_message[0]);
@@ -661,10 +702,10 @@ check_mail_imap(Mailbox *mbox)
 
 	/* Is the machine we are connected to really a IMAP server?
 	*/
-	if (! tcp_getline(mbox, fd, "* OK"))
+	if (! server_response(mbox, fd, "* OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[1]);
 
-	snprintf (line, sizeof (line), ". LOGIN %s %s\r\n", 
+	snprintf (line, sizeof (line), "a001 LOGIN %s %s\r\n", 
 				mbox->username, mbox->password);
 	write(fd, line, strlen (line));
 	if (GK.debug_level & DEBUG_MAIL)
@@ -673,15 +714,17 @@ check_mail_imap(Mailbox *mbox)
 
 		hide_password(mbox, line, strlen(line) - strlen(mbox->password) - 2);
 		format_remote_mbox_name(mbox, buf, sizeof(buf));
-		printf("tcp_sendline( %s ):%s", buf, line);
+		printf("server_command( %s ):%s", buf, line);
 		}
-	if (! tcp_getline(mbox, fd, ". OK"))
+	if (! imap_completion_result(mbox, fd, "a001 OK"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[2]);
 
-	snprintf (line, sizeof (line), ". STATUS %s (MESSAGES UNSEEN)\r\n",
+	/* I expect the only untagged response to STATUS will be "* STATUS ..."
+	*/
+	snprintf (line, sizeof (line), "a002 STATUS %s (MESSAGES UNSEEN)\r\n",
 			mbox->folder);
-	tcp_sendline(mbox, fd, line);
-	if (! tcp_getline(mbox, fd, "*"))
+	server_command(mbox, fd, line);
+	if (! server_response(mbox, fd, "*"))
 		return tcp_shutdown(fd, mbox, tcp_error_message[4]);
 
 	ss = strstr(mbox->tcp_in->str, "MESSAGES");
@@ -694,17 +737,19 @@ check_mail_imap(Mailbox *mbox)
 		}
 	mbox->mail_count = messages;
 	mbox->new_mail_count = unseen;
+	imap_completion_result(mbox, fd, "a002 OK");
+
+	server_command(mbox, fd, "a003 LOGOUT\r\n");
+	imap_completion_result(mbox, fd, "a003 OK");
 
-	tcp_sendline(mbox, fd, ". LOGOUT\r\n");
 	close (fd);
-
 	return TRUE;
 	}
 #endif	/* GKRELLM_THREADS */
 
 
   /* Check a mh directory for mail. The way that messages are marked as new
-  |  depends on the MUA the user is using.  Only .mh_sequences is checked for
+  |  depends on the MUA being using.  Only .mh_sequences is checked for
   |  now, otherwise all mail found is considered new mail.
   */
 static gboolean
@@ -1149,7 +1194,6 @@ fgets_pipe(gchar *line, gint len, Mailproc *mp)
 	return 0;
 	}
 
-
 static gboolean
 mua_is_launched()
 	{
@@ -1160,6 +1204,12 @@ mua_is_launched()
 		return FALSE;
 	while ((n = fgets_pipe(buf, sizeof(buf), &mail_user_agent)) > 0)
 		;
+	if (n == 0 && mail_user_agent.pipe && feof(mail_user_agent.pipe))
+		{
+		pclose(mail_user_agent.pipe);
+		mail_user_agent.pipe = NULL;
+		n = -1;
+		}
 	return (n < 0) ? FALSE : TRUE;
 	}
 
@@ -1727,21 +1777,26 @@ cb_panel_press(GtkWidget *widget, GdkEventButton *ev)
 static Mailbox *
 add_mailbox(gchar *arg)
 	{
-	Mailbox		*mbox;
-	gchar		parms[4][CFG_BUFSIZE];
-	gint		n;
+	Mailbox	*mbox;
+	gchar	parms[4][CFG_BUFSIZE];
+	gint	n;
+	gchar	**str_array = NULL;
 
 	mbox = g_new0(Mailbox, 1);
 	n = sscanf(arg, "%s %s %s %s", parms[0], parms[1], parms[2], parms[3]);
 	switch (n)
 		{
 		case 1:
-			if (isdir(parms[0], "new"))
+			if (parms[0][0] == '~')
+				mbox->path = g_strdup_printf("%s%s", homedir(), parms[0] + 1);
+			else
+				mbox->path = g_strdup(parms[0]);
+			if (isdir(mbox->path, "new"))
 				{
 				mbox->type = MBOX_MAILDIR;
 				mbox->check_func = check_maildir;
 				}
-			else if (isdir(parms[0], NULL))
+			else if (isdir(mbox->path, NULL))
 				{
 				mbox->type = MBOX_MH_DIR;
 				mbox->check_func = check_mh_dir;
@@ -1752,13 +1807,15 @@ add_mailbox(gchar *arg)
 				mbox->type = MBOX_MAILBOX;
 				mbox->check_func = check_mailbox;
 				}
-			mbox->path = g_strdup(parms[0]);
 			break;
 
 #if defined (GKRELLM_THREADS)
 		case 3:
 			mbox->type = MBOX_POP3;
-			mbox->server = g_strdup(parms[0]);
+			str_array = g_strsplit(parms[0], ":", 2);
+			mbox->server = g_strdup(str_array[0]);
+			mbox->port = str_array[1] ? atoi(str_array[1]) : DEFAULT_POP3_PORT;
+			g_strfreev(str_array);
 			mbox->username = g_strdup(parms[1]);
 			mbox->password = g_strdup(parms[2]);
 			mbox->check_func = check_mail_pop3;
@@ -1766,7 +1823,10 @@ add_mailbox(gchar *arg)
 
 		case 4:
 			mbox->type = MBOX_IMAP;
-			mbox->server = g_strdup(parms[0]);
+			str_array = g_strsplit(parms[0], ":", 2);
+			mbox->server = g_strdup(str_array[0]);
+			mbox->port = str_array[1] ? atoi(str_array[1]) : DEFAULT_IMAP_PORT;
+			g_strfreev(str_array);
 			mbox->folder = g_strdup(parms[1]);
 			mbox->username = g_strdup(parms[2]);
 			mbox->password = g_strdup(parms[3]);
@@ -1908,12 +1968,14 @@ save_mail_config(FILE *f)
 				fprintf(f, "mail mailbox %s\n", mbox->path);
 				break;
 			case MBOX_POP3:
-				fprintf(f, "mail mailbox %s %s %s\n", mbox->server,
-					mbox->username, mbox->password);
+				fprintf(f, "mail mailbox %s %s %s\n",
+						format_server_name(mbox),
+						mbox->username, mbox->password);
 				break;
 			case MBOX_IMAP:
-				fprintf(f, "mail mailbox %s %s %s %s\n", mbox->server,
-					mbox->folder, mbox->username, mbox->password);
+				fprintf(f, "mail mailbox %s %s %s %s\n",
+						format_server_name(mbox),
+						mbox->folder, mbox->username, mbox->password);
 				break;
 			default:
 				break;
@@ -2210,6 +2272,7 @@ N_("\nSo, a simple example for monitoring your local mbox and a POP3 mailbox:\n"
 ),
 N_("<i>\t/var/mail/bill\n"),
 N_("<i>\tmail.gnu.org   bill2   PASSWORD\n"),
+
 #else
 N_("Mailbox files to monitor can be mbox or maildir style mailboxes.\n"),
 #endif
@@ -2423,10 +2486,12 @@ _("Note: remote check interval applies to the \"Mail fetch/check program\"\n"
 				*s = '*';
 		password_list = g_list_append(password_list, g_strdup(mbox->password));
 		if (mbox->type == MBOX_POP3)
-			buf[0] = g_strdup_printf("%s   %s   %s", mbox->server,
+			buf[0] = g_strdup_printf("%s   %s   %s",
+					format_server_name(mbox),
 					mbox->username, hidden_passwd);
 		else if (mbox->type == MBOX_IMAP)
-			buf[0] = g_strdup_printf("%s   %s   %s   %s", mbox->server,
+			buf[0] = g_strdup_printf("%s   %s   %s   %s",
+					format_server_name(mbox),
 					mbox->folder, mbox->username, hidden_passwd);
 		else
 			buf[0] = g_strdup(mbox->path);
