@@ -719,7 +719,7 @@ static MeminfoMeter		*mon_in_motion;
 
 static gint			chart_height	= 20;
 
-#define MEG(x)	(((x) + (2 << 19)) >> 20)
+#define MEG(x)	(((x) + (1 << 19)) >> 20)
 
 
   /* Updating the /proc/meminfo meters can be expensive in Linux 2.2,
@@ -732,21 +732,24 @@ static gint			chart_height	= 20;
 #define PIPE_SIZE	3
 static gint		mem_pipe[PIPE_SIZE],
 				swap_pipe[PIPE_SIZE];
+static gint		force_update	= TRUE;
 
-gint
+gboolean
 force_meminfo_update()
 	{
-	gint	i;
+	gint	i, force;
 
+	force = force_update ? TRUE : FALSE;
+	force_update = FALSE;
 	if (GK.second_tick)
 		{
 		for (i = 1; i < PIPE_SIZE; ++i)
 			if (mem_pipe[i] || swap_pipe[i])
-				return TRUE;
+				force = TRUE;
 		if (GK.cpu_sys_activity > 3)
-			return TRUE;
+			force = TRUE;
 		}
-	return FALSE;
+	return force;
 	}
 
 static gint
@@ -1169,6 +1172,7 @@ create_panel(GtkWidget *vbox, MeminfoMeter *mm, gint first_create)
 	TextStyle	*ts;
 	Style		*style;
 
+	(*read_system_meminfo)();
 	if (first_create)
 		mm->panel = gkrellm_panel_new0();
 	else
